
/*
  Copyright (C) 2002-2008  Dmitry V. Levin <ldv@altlinux.org>

  The RPM-dependent stuff for packageof.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <cstdio>
#include <cstring>
#include <cerrno>
#include <error.h>
#include <fcntl.h>
#include <climits>
#include <rpmlib.h>
#include <rpmmacro.h>

#include "rpmdbq.h"

extern int verbose;

RPMDBQ::RPMDBQ(const char *prefix)
{
	if (rpmReadConfigFiles(0, 0))
		error(EXIT_FAILURE, errno,
		      "RPMDBQ: rpmReadConfigFiles failed");

	db = 0;
	if (rpmdbOpen(prefix ? : "/", &db, O_RDONLY, 0644))
		error(EXIT_FAILURE, errno, "RPMDBQ: rpmdbOpen failed");
}

RPMDBQ::~RPMDBQ()
{
	rpmdbClose(db);
	db = 0;
}

int
RPMDBQ::run(PkgMapType & packages, const char *arg)
{
#ifdef DEBUG
	error(0, 0, "run: <arg=%s>", arg);
#endif /* DEBUG */

	if ('/' != arg[0])
	{
		if (verbose)
			error(EXIT_SUCCESS, 0,
			      "file name \"%s\" is not absolute", arg);
		return EXIT_FAILURE;
	}

	char    fn[PATH_MAX];

	if (strlcpy(fn, arg, sizeof(fn)) >= sizeof(fn))
	{
		if (verbose)
			error(EXIT_SUCCESS, 0,
			      "file name \"%s\" is too long", arg);
		return EXIT_FAILURE;
	}

	const char *fname = rpmCleanPath(fn);

#ifdef DEBUG
	error(0, 0, "run: <fname=%s>", fname);
#endif /* DEBUG */

	if (!fname || !fname[0])
	{
		if (verbose)
			error(EXIT_SUCCESS, 0,
			      "file name \"%s\" is not acceptable", arg);
		return EXIT_FAILURE;
	}

	struct stat st;

	if (stat(fname, &st))
	{
		if (verbose)
			error(EXIT_SUCCESS, errno,
			      "file \"%s\" is not available", arg);
		return EXIT_FAILURE;
	}

	if (!S_ISREG(st.st_mode))
	{
		if (verbose)
			error(EXIT_SUCCESS, errno,
			      "non-regular file \"%s\", ignoring", arg);
		return EXIT_SUCCESS;
	}

	rpmdbMatchIterator mi;

	if (!(mi = rpmdbInitIterator((rpmdb) db, RPMTAG_BASENAMES, fn, 0)))
	{
		if (verbose)
			error(EXIT_SUCCESS, 0,
			      "file \"%s\" is not owned by any package", fn);
		return EXIT_FAILURE;
	}

	Header  h;
	int     rc = EXIT_SUCCESS;

	while ((h = rpmdbNextIterator(mi)))
	{
		void   *s_void = 0;
		const char *s;
		int     st = headerGetEntry(h, RPMTAG_NAME, 0, &s_void, 0);

		s = (const char *) s_void;
#ifdef DEBUG
		error(0, 0, "run: h=%p, rc=%d, t=%u, s=%s, c=%u",
		      h, rc, t, s, c);
#endif /* DEBUG */

		if (st && s)
			packages[s] = 1;
		else
			rc = EXIT_FAILURE;
	}

	mi = rpmdbFreeIterator(mi);
	return rc;
}
